@@ -54,6 +54,8 @@ class Session(object):
         self.expiry_time = None
         self.client = None
         self.maximum_packet_size = MAXIMUM_PACKET_SIZE
+        self.will_topic = None
+        self.will_message = None
 
     def clean(self):
         self.subscriptions = set()
@@ -61,6 +63,8 @@ class Session(object):
         self.expiry_time = None
         self.client = None
         self.maximum_packet_size = MAXIMUM_PACKET_SIZE
+        self.will_topic = None
+        self.will_message = None
 
 
 def is_wildcards_in_topic(topic):
@@ -102,11 +106,17 @@ class Client(object):
             elif isinstance(e, ProtocolError):
                 self._disconnect_reason = DisconnectReasonCode.PROTOCOL_ERROR
 
-            self.disconnect()
+            if self._session is not None:
+                self.disconnect()
 
         if self._session is not None:
             self._session.client = None
 
+            # ToDo: Should not publish on normal disconnection.
+            if self._session.will_topic is not None:
+                self._broker.publish(self._session.will_topic,
+                                     self._session.will_message)
+
         self.log_info('Closing client %r.', addr)
 
     async def reader_loop(self):
@@ -150,6 +160,8 @@ class Client(object):
     def on_connect(self, payload):
         (client_id,
          clean_start,
+         will_topic,
+         will_message,
          keep_alive_s,
          properties,
          user_name,
@@ -175,6 +187,9 @@ class Client(object):
             maximum_packet_size = properties[PropertyIds.MAXIMUM_PACKET_SIZE]
             self._session.maximum_packet_size = maximum_packet_size
 
+        self._session.will_topic = will_topic
+        self._session.will_message = will_message
+
         if (user_name is not None) or (password is not None):
             reason = ConnectReasonCode.BAD_USER_NAME_OR_PASSWORD
 
@@ -199,8 +214,7 @@ class Client(object):
         if is_wildcards_in_topic(topic):
             raise MalformedPacketError(f'Invalid topic {topic} in publish.')
 
-        for session in self._broker.iter_subscribers(topic):
-            session.client.publish(topic, message)
+        self._broker.publish(topic, message)
 
     def on_subscribe(self, payload):
         packet_identifier, _, subscriptions = unpack_subscribe(payload)
@@ -383,3 +397,11 @@ class Broker(object):
             self._sessions[client_id] = session
 
         return session, session_present
+
+    def publish(self, topic, message):
+        """Publish given topic and message to all subscribers.
+
+        """
+
+        for session in self.iter_subscribers(topic):
+            session.client.publish(topic, message)
