@@ -9,7 +9,6 @@ from .common import PropertyIds
 from .common import SubackReasonCode
 from .common import UnsubackReasonCode
 from .common import DisconnectReasonCode
-from .common import QoS
 from .common import pack_connect
 from .common import unpack_connack
 from .common import pack_disconnect
@@ -121,6 +120,15 @@ class Client(object):
     seconds. Give as 0 to remove the session in the broker when the
     connection ends.
 
+    `subscriptions` is a list of topics to subscribe to after
+    connected in :meth:`start()`.
+
+    `connect_delays` is a list of delays in seconds between the
+    connection attempts in :meth:`start()`. Each delay is used once,
+    except the last delay, which is used until successfully
+    connected. If ``[]``, only one connection attempt is performed. If
+    ``None``, the default delays ``[1, 2, 4, 8]`` are used.
+
     `kwargs` are passed to ``asyncio.open_connection()``.
 
     Create a client with default configuration:
@@ -134,12 +142,14 @@ class Client(object):
                         client_id='my-client',
                         will_topic='/my/last/will',
                         will_message=b'my-last-message',
-                        will_qos=QoS.EXACTLY_ONCE,
+                        will_qos=0,
                         keep_alive_s=600,
                         response_timeout=30',
                         topic_aliases=['/my/topic']',
                         topic_alias_maximum=100,
                         session_expiry_interval=1800,
+                        subscriptions=['a/b', 'test/#'],
+                        connect_delays=[1, 2],
                         ssl=True)
 
     """
@@ -150,12 +160,14 @@ class Client(object):
                  client_id=None,
                  will_topic='',
                  will_message=b'',
-                 will_qos=QoS.AT_MOST_ONCE,
+                 will_qos=0,
                  keep_alive_s=60,
                  response_timeout=5,
                  topic_aliases=None,
                  topic_alias_maximum=10,
                  session_expiry_interval=0,
+                 subscriptions=None,
+                 connect_delays=None,
                  **kwargs):
         self._host = host
         self._port = port
@@ -182,6 +194,16 @@ class Client(object):
             self._connect_properties[PropertyIds.SESSION_EXPIRY_INTERVAL] = (
                 session_expiry_interval)
 
+        if subscriptions is None:
+            subscriptions = []
+
+        self._subscriptions = subscriptions
+
+        if connect_delays is None:
+            connect_delays = [1, 2, 4, 8]
+
+        self._connect_delays = connect_delays
+
         if topic_aliases is None:
             topic_aliases = []
 
@@ -235,6 +257,15 @@ class Client(object):
 
         return self._messages
 
+    async def on_message(self, topic, message):
+        """Called for each received MQTT message and when the broker
+        connection is lost. Puts the message on the messages queue by
+        default.
+
+        """
+
+        await self._messages.put((topic, message))
+
     async def start(self, resume_session=False):
         """Open a TCP connection to the broker and perform the MQTT connect
         procedure. This method must be called before any
@@ -246,6 +277,9 @@ class Client(object):
         exception is raised if the resume fails, and a new session has
         been created instead.
 
+        The exceptions below are only raised if ``connect_delays`` is
+        ``[]``.
+
         Raises ``ConnectionRefusedError`` if the TCP connection
         attempt is refused by the broker.
 
@@ -255,6 +289,9 @@ class Client(object):
         Raises :class:`ConnectError` if the broker does not accept the
         connect request.
 
+        Raises :class:`SubscribeError` if the broker does not accept
+        the subscribe request(s).
+
         >>> await client.start()
 
         Trying to resume a session.
@@ -267,10 +304,43 @@ class Client(object):
 
         """
 
+        attempt = 1
+        delays = self._connect_delays
+
+        while True:
+            try:
+                await self._start(resume_session)
+                break
+            except Exception as e:
+                if isinstance(e, ConnectionRefusedError):
+                    LOGGER.info('TCP connect refused.')
+                elif isinstance(e, TimeoutError):
+                    LOGGER.info(
+                        'MQTT connect or subscribe acknowledge not received.')
+                elif isinstance(e, ConnectError):
+                    LOGGER.info('MQTT connect failed with reason %s.', e)
+                elif isinstance(e, SubscribeError):
+                    LOGGER.info('MQTT subscribe failed with reason %s.', e)
+
+                if delays == []:
+                    raise
+
+            # Delay a while before the next connect attempt.
+            delay = delays[min(attempt, len(delays)) - 1]
+            LOGGER.info(
+                'Waiting %s second(s) before next connection attempt(%d).',
+                delay,
+                attempt)
+            await asyncio.sleep(delay)
+            attempt += 1
+
+    async def _start(self, resume_session=False):
         self._rx_topic_aliases = {}
         self._tx_topic_aliases = {}
         self._tx_topic_alias_maximum = 0
         self._registered_broker_topic_aliases = set()
+        self._reader_task = None
+        self._keep_alive_task = None
         self._connack_event = asyncio.Event()
         self._pingresp_event = asyncio.Event()
         self.transactions = {}
@@ -281,18 +351,25 @@ class Client(object):
 
         LOGGER.info('Connecting to %s:%s.', self._host, self._port)
 
-        self._reader, self._writer = await asyncio.open_connection(
-            self._host,
-            self._port,
-            **self._kwargs)
-        self._reader_task = asyncio.create_task(self._reader_main())
+        try:
+            self._reader, self._writer = await asyncio.open_connection(
+                self._host,
+                self._port,
+                **self._kwargs)
 
-        if self._keep_alive_s != 0:
-            self._keep_alive_task = asyncio.create_task(self._keep_alive_main())
-        else:
-            self._keep_alive_task = None
+            self._reader_task = asyncio.create_task(self._reader_main())
 
-        await self.connect(resume_session)
+            await self.connect(resume_session)
+
+            if self._keep_alive_s != 0:
+                self._keep_alive_task = asyncio.create_task(self._keep_alive_main())
+
+            if not resume_session:
+                for topic in self._subscriptions:
+                    await self.subscribe(topic)
+        except Exception:
+            await self.stop()
+            raise
 
     async def stop(self):
         """Try to cleanly disconnect from the broker and then close the TCP
@@ -308,12 +385,13 @@ class Client(object):
         except Exception:
             pass
 
-        self._reader_task.cancel()
+        if self._reader_task is not None:
+            self._reader_task.cancel()
 
-        try:
-            await self._reader_task
-        except Exception:
-            pass
+            try:
+                await self._reader_task
+            except Exception:
+                pass
 
         if self._keep_alive_task is not None:
             self._keep_alive_task.cancel()
@@ -323,7 +401,8 @@ class Client(object):
             except Exception:
                 pass
 
-        self._writer.close()
+        if self._writer is not None:
+            self._writer.close()
 
     async def connect(self, resume_session):
         self._connack_event.clear()
@@ -480,7 +559,7 @@ class Client(object):
                              alias)
                 return
 
-        await self._messages.put((topic, message))
+        await self.on_message(topic, message)
 
     def on_suback(self, payload):
         packet_identifier, properties, reasons = unpack_suback(payload)
@@ -542,6 +621,8 @@ class Client(object):
 
         """
 
+        LOGGER.info('Reader task starting...')
+
         try:
             await self.reader_loop()
         except Exception as e:
@@ -568,6 +649,8 @@ class Client(object):
 
         """
 
+        LOGGER.info('Keep alive task starting...')
+
         try:
             await self.keep_alive_loop()
         except Exception as e:
@@ -618,4 +701,4 @@ class Client(object):
     async def _close(self):
         self.disconnect()
         self._writer.close()
-        await self._messages.put((None, None))
+        await self.on_message(None, None)
