@@ -3,7 +3,7 @@ import struct
 import enum
 from io import BytesIO
 import bitstruct
-from binascii import hexlify
+import binascii
 
 
 # Connection flags.
@@ -16,6 +16,10 @@ PASSWORD_FLAG  = 0x40
 USER_NAME_FLAG = 0x80
 
 
+def hexlify(data):
+    return binascii.hexlify(data).decode('ascii').upper()
+
+
 # Control packet types.
 class ControlPacketType(enum.IntEnum):
     CONNECT     = 1
@@ -97,40 +101,6 @@ class DisconnectReasonCode(enum.IntEnum):
     WILDCARD_SUBSCRIPTIONS_NOT_SUPPORTED   = 162
 
 
-class PubackReasonCode(enum.IntEnum):
-    SUCCESS                       = 0
-    NO_MATCHING_SUBSCRIBERS       = 16
-    UNSPECIFIED_ERROR             = 128
-    IMPLEMENTATION_SPECIFIC_ERROR = 131
-    NOT_AUTHORIZED                = 135
-    TOPIC_NAME_INVALID            = 144
-    PACKET_IDENTIFIER_IN_USE      = 145
-    QUOTA_EXCEEDED                = 151
-    PAYLOAD_FORMAT_INVALID        = 153
-
-
-class PubrecReasonCode(enum.IntEnum):
-    SUCCESS                       = 0
-    NO_MATCHING_SUBSCRIBERS       = 16
-    UNSPECIFIED_ERROR             = 128
-    IMPLEMENTATION_SPECIFIC_ERROR = 131
-    NOT_AUTHORIZED                = 135
-    TOPIC_NAME_INVALID            = 144
-    PACKET_IDENTIFIER_IN_USE      = 145
-    QUOTA_EXCEEDED                = 151
-    PAYLOAD_FORMAT_INVALID        = 153
-
-
-class PubrelReasonCode(enum.IntEnum):
-    SUCCESS                     = 0
-    PACKET_IDENTIFIER_NOT_FOUND = 146
-
-
-class PubcompReasonCode(enum.IntEnum):
-    SUCCESS                     = 0
-    PACKET_IDENTIFIER_NOT_FOUND = 146
-
-
 class SubackReasonCode(enum.IntEnum):
     GRANTED_QOS_0                          = 0
     GRANTED_QOS_1                          = 1
@@ -186,21 +156,6 @@ class PropertyIds(enum.IntEnum):
     SHARED_SUBSCRIPTION_AVAILABLE     = 42
 
 
-class QoS(enum.IntEnum):
-    """Quality of service levels.
-
-    """
-
-    AT_MOST_ONCE  = 0
-    "At most once (QoS level 0)."
-
-    AT_LEAST_ONCE = 1
-    "At least once (QoS level 1)."
-
-    EXACTLY_ONCE  = 2
-    "Exactly once (QoS level 2)."
-
-
 # MQTT 5.0
 PROTOCOL_VERSION = 5
 
@@ -458,7 +413,7 @@ def pack_connect(client_id,
 
     payload_length = len(client_id) + 2
 
-    if will_topic and will_message:
+    if will_topic:
         flags |= WILL_FLAG
 
         if will_qos == 1:
@@ -475,8 +430,7 @@ def pack_connect(client_id,
     packed = pack_fixed_header(ControlPacketType.CONNECT,
                                0,
                                10 + payload_length + len(properties))
-    packed += struct.pack('>H', 4)
-    packed += b'MQTT'
+    packed += pack_string('MQTT')
     packed += struct.pack('>BBH',
                           PROTOCOL_VERSION,
                           flags,
@@ -493,9 +447,7 @@ def pack_connect(client_id,
 
 
 def unpack_connect(payload):
-    unpack_u16(payload)
-
-    if payload.read(4) != b'MQTT':
+    if unpack_string(payload) != 'MQTT':
         raise MalformedPacketError('Invalid MQTT magic string.')
 
     if unpack_u8(payload) != PROTOCOL_VERSION:
@@ -521,7 +473,22 @@ def unpack_connect(payload):
     client_id = unpack_string(payload)
 
     if flags & WILL_FLAG:
-        raise MalformedPacketError()
+        will_properties = unpack_properties(
+            'CONNECT-WILL',
+            [
+                PropertyIds.WILL_DELAY_INTERVAL,
+                PropertyIds.PAYLOAD_FORMAT_INDICATOR,
+                PropertyIds.MESSAGE_EXPIRY_INTERVAL,
+                PropertyIds.CONTENT_TYPE,
+                PropertyIds.RESPONSE_TOPIC,
+                PropertyIds.USER_PROPERTY
+            ],
+            payload)
+        will_topic = unpack_string(payload)
+        will_message = unpack_binary(payload)
+    else:
+        will_topic = None
+        will_message = None
 
     if flags & USER_NAME_FLAG:
         user_name = unpack_string(payload)
@@ -533,7 +500,14 @@ def unpack_connect(payload):
     else:
         password = None
 
-    return client_id, clean_start, keep_alive_s, properties, user_name, password
+    return (client_id,
+            clean_start,
+            will_topic,
+            will_message,
+            keep_alive_s,
+            properties,
+            user_name,
+            password)
 
 
 def pack_connack(session_present,
@@ -800,17 +774,21 @@ def format_properties(properties):
 def format_connect(payload):
     (client_id,
      clean_start,
+     will_topic,
+     will_message,
      keep_alive_s,
      properties,
      user_name,
      password) = unpack_connect(payload)
 
     return [
-        f'  ClientId:   {client_id}',
-        f'  CleanStart: {clean_start}',
-        f'  KeepAlive:  {keep_alive_s}',
-        f'  UserName:   {user_name}',
-        f'  Password:   {password}'
+        f'  ClientId:    {client_id}',
+        f'  CleanStart:  {clean_start}',
+        f'  WillTopic:   {will_topic}',
+        f'  WillMessage: {hexlify(will_message)}',
+        f'  KeepAlive:   {keep_alive_s}',
+        f'  UserName:    {user_name}',
+        f'  Password:    {password}'
     ] + format_properties(properties)
 
 
@@ -834,7 +812,7 @@ def format_publish(flags, payload):
         f'  QoSLevel:   {qos}',
         f'  Retain:     {retain}',
         f'  Topic:      {topic}',
-        f'  Message:    {message}',
+        f'  Message:    {hexlify(message)}',
         '  Properties:'
     ] + format_properties(properties)
 
