@@ -1,11 +1,20 @@
 import asyncio
 import time
+import binascii
 from humanfriendly import format_timespan
 
 from ..client import Client
+from ..common import Error
 
 
-def create_message_bytes(message, size, number, fmt):
+def encode_message(message):
+    try:
+        return binascii.unhexlify(message)
+    except ValueError:
+        raise Error(f"Invalid hex string '{message}'.")
+
+
+def create_message(message, size, number, fmt):
     if message is None:
         message_bytes = fmt.format(number).encode('ascii')
         extra = (size - len(message_bytes))
@@ -15,13 +24,28 @@ def create_message_bytes(message, size, number, fmt):
         else:
             message_bytes = message_bytes[:size]
     else:
-        message_bytes = message.encode('ascii')
+        message_bytes = encode_message(message)
 
     return message_bytes
 
 
-async def publisher(host, port, client_id, count, size, topic, message):
-    client = Client(host, port, client_id)
+async def publisher(host,
+                    port,
+                    client_id,
+                    count,
+                    size,
+                    will_topic,
+                    will_message,
+                    topic,
+                    message):
+    if will_message is not None:
+        will_message = encode_message(will_message)
+
+    client = Client(host,
+                    port,
+                    client_id,
+                    will_topic=will_topic,
+                    will_message=will_message)
 
     print(f"Connecting to '{host}:{port}'.")
     print()
@@ -32,7 +56,7 @@ async def publisher(host, port, client_id, count, size, topic, message):
     start_time = time.time()
 
     for number in range(count):
-        message_bytes = create_message_bytes(message, size, number, fmt)
+        message_bytes = create_message(message, size, number, fmt)
         client.publish(topic, message_bytes)
 
     elapsed_time = format_timespan(time.time() - start_time)
@@ -47,6 +71,8 @@ def _do_publish(args):
                           args.client_id,
                           args.count,
                           args.size,
+                          args.will_topic,
+                          args.will_message,
                           args.topic,
                           args.message))
 
@@ -55,7 +81,7 @@ def add_subparser(subparsers):
     subparser = subparsers.add_parser('publish',
                                       description='Publish given topic.')
     subparser.add_argument('--host',
-                           default='broker.hivemq.com',
+                           default='localhost',
                            help='Broker host (default: %(default)s).')
     subparser.add_argument('--port',
                            type=int,
@@ -73,6 +99,8 @@ def add_subparser(subparsers):
         type=int,
         default=10,
         help='Generated message size (default: %(default)s).')
+    subparser.add_argument('--will-topic', help='Will topic.')
+    subparser.add_argument('--will-message', help='Will message.')
     subparser.add_argument('topic', help='Topic to publish.')
     subparser.add_argument(
         'message',
